@@ -46,25 +46,37 @@ ESSENTIA_D_VERSION = re.compile('MPU_E6D[A-Za-z\t .]+')
 #MPU_A4D_vx.yy
 '''
 SIMPLESE_VERSION = re.compile('MPU_A4D[A-Za-z\t .]+')
+'''
+#ALLOFF
+'''
+ALL_OFF_PATTERN = re.compile('ALLOFF')
+'''
+#?
+'''
+ERROR_PATTERN = re.compile('\?')
+
 
 EOL = b'\r'
 TIMEOUT_OP       = 0.2   # Number of seconds before serial operation timeout
 TIMEOUT_RESPONSE = 1     # Number of seconds before command response timeout
-zoneinit = np.array([0,0,0,0,0,0,0,0,0,0,0,0,0])
+zoneinit = np.array([1,0,0,0,0,0,0,0,0,0,0,0,0])
 zonesetinit = np.array([0,0,0,0,0,0,0,0,0,0,0,0,0])
 zonepwr = np.array([0,0,0,0,0,0,0,0,0,0,0,0,0])
 zonepwrsave = np.array([0,0,0,0,0,0,0,0,0,0,0,0,0])
-zonevol = np.array([0,-78,-78,-78,-78,-78,-78,-78,-78,-78,-78,-78,-78])
-zonevolsave = np.array([0,-78,-78,-78,-78,-78,-78,-78,-78,-78,-78,-78,-78])
-zonesrc = np.array([0,0,0,0,0,0,0,0,0,0,0,0,0])
+zonepwralloff = np.array([0,0,0,0,0,0,0,0,0,0,0,0,0])
+zonevol = np.array([-78,-78,-78,-78,-78,-78,-78,-78,-78,-78,-78,-78,-78])
+zonevolsave = np.array([-78,-78,-78,-78,-78,-78,-78,-78,-78,-78,-78,-78,-78])
+zonesrc = np.array([1,0,0,0,0,0,0,0,0,0,0,0,0])
 zonesrcsave = np.array([0,1,1,1,1,1,1,1,1,1,1,1,1])
 zonemute = np.array([0,0,0,0,0,0,0,0,0,0,0,0,0])
 zonemutesave = np.array([0,0,0,0,0,0,0,0,0,0,0,0,0])
 zonebass = np.array([0,0,0,0,0,0,0,0,0,0,0,0,0])
 zonetreble = np.array([0,0,0,0,0,0,0,0,0,0,0,0,0])
-zonegroup = np.array([0,1,1,1,1,1,1,1,1,1,1,1,1])
+zonegroupsrc = np.array([1,1,1,1,1,1,1,1,1,1,1,1,1])
 zoneoverride = np.array([0,0,0,0,0,0,0,0,0,0,0,0,0])
 zonevolreset = np.array([0,1,1,1,1,1,1,1,1,1,1,1,1])
+zonevoloffset = np.array([0,0,0,0,0,0,0,0,0,0,0,0,0])
+setup = False
 
 class ZoneStatus(object):
     def __init__(self
@@ -75,25 +87,33 @@ class ZoneStatus(object):
                  ,volume: float  # -78 -> 0
                  ,from_port: bool
                  ):
-
         self.source = str(source)
         self.zone = abs(int(zone))
+        self.zonegroup_members = []
+        try:
+            zoneindexobj = [i for i, x in enumerate(media_zonetable) if x == zone]
+            zoneindex = int(zoneindexobj[0])
+            for grpzonembr in groupzone_members[zoneindex]:
+                entityindexobj = [i for i, x in enumerate(media_zonetable) if x == grpzonembr]
+                entityindex = int(entityindexobj[0])
+                self.zonegroup_members.append(entitytable[entityindex])
+        except:
+            self.zonegroup_members = None
         if 'ON' in power:
            self.power = bool(1)
            """ Nuvo does not send update from other zones using source
                grouping.  See if is enabled and if so, update the other
-               zones that are powered up. """
-           if zonegroup[self.zone] == 0:
+               zones that are powered on. """
+           if zonegroupsrc[self.zone] == 0:
                for count in range(1,13):
-                   if zonegroup[count] == 0 and zonepwr[count] == 1:
-                       zonesrc[count] = self.source
-                       if from_port:
-                           try:
-                               callbackobj = media_zonetable.index(count)
-                               callbackint = int(callbackobj)
-                               media_callback[callbackint]()
-                           except:
-                               callbackobj =  None
+                   if zonegroupsrc[count] == 0 and zonepwr[count] == 1:
+                       if int(zonesrc[count]) != int(self.source):
+                           zonesrc[count] = self.source
+                           if from_port:
+                               if zone in media_zonetable:
+                                   callbackobj = media_zonetable.index(count)
+                                   callbackint = int(callbackobj)
+                                   media_callback[callbackint]()
 
         else:
            self.power = bool(0)
@@ -111,16 +131,14 @@ class ZoneStatus(object):
         zonemute[self.zone] = self.mute
 
         if from_port:
-            try:
-               callbackobj = media_zonetable.index(self.zone)
-               callbackint = int(callbackobj)
-               media_callback[callbackint]()
-            except:
-               callbackobj =  None
+            if self.zone in media_zonetable:
+                callbackobj = media_zonetable.index(self.zone)
+                callbackint = int(callbackobj)
+                media_callback[callbackint]()
 
-        _LOGGER.debug('Zone: %s, Power: %s, Vol: %s, Mute: %s, Source: %s, Init: %s', \
-           self.zone, self.power, self.volume, self.mute, self.source, \
-           zoneinit[int(self.zone)])
+        _LOGGER.debug('Zone: %s, Power: %s, Vol: %sdb, %s%%, Mute: %s, Source: %s, Init: %s', \
+           self.zone, self.power, self.volume, db_to_percent(self.volume), \
+           self.mute, self.source, zoneinit[int(self.zone)])
 
     @classmethod
     def from_string(cls, match):
@@ -142,10 +160,12 @@ class ZonesetStatus(object):
         self.bass = bass
         zonebass[self.zone] = self.bass
         zonetreble[self.zone] = self.treble
+        self.volume_offset = zonevoloffset[self.zone]
+
         if MODEL == 'ESSENTIA_D':
             zoneoverride[self.zone] = int(ovride)
         zonevolreset[self.zone] = vrset
-        zonegroup[self.zone] = grp
+        zonegroupsrc[self.zone] = grp
 
         if zoneoverride[self.zone] == 1:
            self.override = bool(1)
@@ -155,18 +175,17 @@ class ZonesetStatus(object):
            self.volume_reset = bool(0)
         else:
            self.volume_reset = bool(1)
-        if zonegroup[self.zone] == 1:
+        if zonegroupsrc[self.zone] == 1:
            self.group = bool(0)
         else:
            self.group = bool(1)
 
         if from_port:
-            try:
-                callbackindex = [i for i, x in enumerate(settings_zonetable) if x == self.zone]
-                for callbackzone in callbackindex:
-                    settings_callback[callbackzone]()
-            except:
-               callbackobj =  None
+            if 'settings_zonetable' in globals():
+                if self.zone in settings_zonetable:
+                    callbackindex = [i for i, x in enumerate(settings_zonetable) if x == self.zone]
+                    for callbackzone in callbackindex:
+                        settings_callback[callbackzone]()
 
         _LOGGER.debug('Zone: %s, Bass: %s, Treble: %s, Override: %s, VolReset: %s, Group: %s, Init: %s',\
             self.zone, self.bass, self.treble, self.override, \
@@ -183,7 +202,7 @@ class Nuvo(object):
     """
     Nuvo amplifier interface
     """
-    def add_callback(self, coro: Callable[..., Coroutine], zone) -> None:
+    def add_callback(self, coro: Callable[..., Coroutine], zone, zone_name) -> None:
         """
         Add entity subscription for updates
         """
@@ -219,6 +238,21 @@ class Nuvo(object):
         """
         raise NotImplemented()
 
+    def join_players(self, zone, group_members):
+        """
+        Join entity to a speaker group
+        :param zone: zone 1.12
+        :param group_members: entity to add
+        """
+        raise NotImplemented()
+
+    def unjoin_player(self, zone):
+        """
+        Removes zone from all speaker groups
+        :param zone to remove
+        """
+        raise NotImplemented()
+
     def set_mute(self, zone: int, mute: bool):
         """
         Mute zone on or off
@@ -235,6 +269,14 @@ class Nuvo(object):
         """
         raise NotImplemented()
 
+    def set_volume_offset(self, zone: int, volume: float):
+        """
+        Set volume offset for groups
+        :param zone: zone 1.12
+        :param volume: float from -30 to 30 inclusive
+        """
+        raise NotImplemented()
+
     def set_treble(self, zone: int, treble: float):
         """
         Set treble for zone
@@ -298,6 +340,18 @@ def _is_int(s):
     except ValueError:
         return False
 
+def db_to_percent(db):
+    if int(db) > -24:
+        return int((( int(db) + 24) / 24 / 2 +.5) * 100)
+    else:
+        return int((( int(db) + 78) / 54 / 2) * 100)
+
+def percent_to_db(percent):
+    if percent > 50:
+        return int(((percent / 100) * 48) - 48)
+    else:
+        return int(((percent / 100) * 108) - 78)
+
 def _parse_response(string: bytes):
    """
    :param request: request that is sent to the nuvo
@@ -320,6 +374,17 @@ def _parse_response(string: bytes):
           _LOGGER.debug('Simplese SETSR received')
           ZonesetStatus.from_string(match)
 
+   if not match:
+       match = re.search(ALL_OFF_PATTERN, string)
+       if match:
+          _LOGGER.debug('ALL OFF received')
+          all_off_keypad()
+
+   if not match:
+       match = re.search(ERROR_PATTERN, string)
+       if match:
+          _LOGGER.error('Received error response from Nuvo on last command')
+
    global MODEL
 
    if not match:
@@ -384,7 +449,6 @@ def _format_set_volume(zone: int, volume: float) -> str:
     # If muted, status has no info on volume level
     if _is_int(volume):
        # Negative sign in volume parm produces erronous result
-       zonevol[int(zone)] = int(volume)
        volume = abs(volume)
        volume = round(volume,0)
     return 'Z{:0=2}VOL{:0=2}'.format(int(zone), volume)
@@ -417,7 +481,25 @@ def _format_unmute_all() -> str:
 def _format_all_off() -> str:
     return 'ALLOFF'.format()
 
-def get_nuvo(port_url, baud):
+def all_off_keypad():
+    time.sleep(0.5) # Nuvo does NOT process commands immediately after ALLOFF
+    zonesalloff = 1
+    for count in range(1,13):
+        if zonepwr[int(count)] == 1:
+            zonesalloff = 0
+    for count in range(1,13):
+        if zonesalloff == 0:
+            zonepwralloff[int(count)] = zonepwr[int(count)]
+            zonepwr[int(count)] = 0
+        else:
+            if zonepwralloff[int(count)] == 1 and ALL_OFF_RECALL:
+                zonepwr[int(count)] = 1
+        if count in media_zonetable:
+            callbackobj = media_zonetable.index(count)
+            callbackint = int(callbackobj)
+            media_callback[callbackint]()
+
+def get_nuvo(port_url, baud, all_off_recall):
     """
     Return synchronous version of Nuvo interface
     :param port_url: serial port, i.e. '/dev/ttyUSB0,/dev/ttyS0'
@@ -436,6 +518,8 @@ def get_nuvo(port_url, baud):
 
     class NuvoSync(Nuvo):
         def __init__(self, port_url):
+            global ALL_OFF_RECALL
+            ALL_OFF_RECALL = all_off_recall
             self._port = serial.serial_for_url(port_url, do_not_open=True)
             self._port.baudrate = int(baud)
             self._port.stopbits = serial.STOPBITS_ONE
@@ -510,11 +594,14 @@ def get_nuvo(port_url, baud):
         SyncThread = threading.Thread(target=_listen, args=(), daemon=True)
         SyncThread.start()
 
-        def add_callback(self, callback: Callable[..., Coroutine], zone, entity) -> None:
+        def add_callback(self, callback: Callable[..., Coroutine], zone, entity_id, entity_type) -> None:
             global media_zonetable
             global media_callback
             global settings_zonetable
             global settings_callback
+            global entitytable
+            global groupzone_members
+            global setup
             try:
                 media_zonetable
             except:
@@ -522,16 +609,25 @@ def get_nuvo(port_url, baud):
                 media_callback = []
                 settings_zonetable = []
                 settings_callback = []
-            if entity == 'media':
+                entitytable = []
+                groupzone_members = []
+            if entity_type == 'media':
                 zone = int(zone)
+                entitytable.append(entity_id)
                 media_zonetable.append(zone)
                 media_callback.append(callback)
-                _LOGGER.debug('Added %s callback: %s', entity, callback)
-            if entity == 'settings':
+                groupzone_members.append(zone)
+                zoneindexobj = [i for i, x in enumerate(media_zonetable) if x == zone]
+                zoneindex = int(zoneindexobj[0])
+                groupzone_members[zoneindex] = []
+                groupzone_members[zoneindex].append(zone)
+                setup = True
+                _LOGGER.debug('Added %s media callback: %s', entity_id, callback)
+            if entity_type == 'settings':
                 zone = int(zone)
                 settings_zonetable.append(zone)
                 settings_callback.append(callback)
-                _LOGGER.debug('Added %s callback: %s', entity, callback)
+                _LOGGER.debug('Added %s settings callback: %s', entity_id, callback)
 
         @synchronized
         def zone_status(self, zone: int):
@@ -557,8 +653,50 @@ def get_nuvo(port_url, baud):
                else:
                   zonestrvol = str(zonevol[int(zone)])
 
-               return ZoneStatus(int(zone), zonestrpwr,
-                                    zonesrc[int(zone)], 0, zonestrvol, False)
+            if zonepwralloff[int(zone)] == 1 and zonepwr[int(zone)] == 1:
+                self.set_power(int(zone), True)
+                zonepwralloff[int(zone)] = 0
+
+            if setup:
+                zoneindexobj = [i for i, x in enumerate(media_zonetable) if x == zone]
+                zoneindex = int(zoneindexobj[0])
+                if len(groupzone_members[zoneindex]) > 1:
+                    zonecnt = 1
+                    while zonecnt < len(groupzone_members[zoneindex]):
+                        group_member = int(groupzone_members[zoneindex][zonecnt])
+                        _LOGGER.debug('Zone %s: update group member zone %s', \
+                            zone, group_member)
+                        if zonepwr[int(zone)] != zonepwr[group_member]:
+                            if zonepwr[int(zone)] == 1:
+                                self.set_power(group_member, True)
+                                time.sleep(0.05) # Give Nuvo zone time to power up
+                            else:
+                                self.set_power(group_member, False)
+                        if zonevol[int(zone)] != zonevol[group_member]:
+                            if zonevoloffset[group_member] != 0:
+                                offsetvolpct = db_to_percent(zonevol[int(zone)]) + \
+                                    zonevoloffset[group_member]
+                                offsetvol = percent_to_db(offsetvolpct)
+                                if offsetvol < -78:
+                                    offsetvol = -78
+                                else:
+                                    if offsetvol > 0:
+                                        offsetvol = 0
+                                self.set_volume(group_member, offsetvol)
+                            else:
+                                self.set_volume(group_member, zonevol[int(zone)])
+                        if zonemute[int(zone)] != zonemute[group_member]:
+                            if zonemute[int(zone)] == 1:
+                                self.set_mute(group_member, True)
+                            else:
+                                self.set_mute(group_member, False)
+                        if zonesrc[int(zone)] != zonesrc[group_member]:
+                            self.set_source(group_member, zonesrc[int(zone)])
+                        zonecnt += 1
+
+            return ZoneStatus(int(zone), zonestrpwr,
+                zonesrc[int(zone)], 0, zonestrvol, False)
+
 
         @synchronized
         def zoneset_status(self, zone: int):
@@ -579,20 +717,95 @@ def get_nuvo(port_url, baud):
                return ZonesetStatus(int(zone), zoneoverride[int(zone)],
                                     zonebass[int(zone)],
                                     zonetreble[int(zone)],
-                                    zonegroup[int(zone)],
+                                    zonegroupsrc[int(zone)],
                                     zonevolreset[int(zone)], False)
 
+        @synchronized
+        def join_players(self, zone, group_members):
+            zoneindexobj = [i for i, x in enumerate(media_zonetable) if x == zone]
+            zoneindex = int(zoneindexobj[0])
+            for group_member in group_members:
+                entityindexobj = [i for i, x in enumerate(entitytable) if x == group_member]
+                entityindex = int(entityindexobj[0])
+                addedzone = int(media_zonetable[entityindex])
+                del groupzone_members[entityindex][:]
+                groupzone_members[entityindex].append(media_zonetable[entityindex])
+                for zones in media_zonetable:
+                    zonesindexobj = [i for i, x in enumerate(media_zonetable) if x == zones]
+                    zonesindex = int(zonesindexobj[0])
+                    if zones != zone:
+                        if zone in groupzone_members[zonesindex]:
+                            groupzone_members[zonesindex].remove(zone)
+                    if zones != addedzone:
+                        if addedzone in groupzone_members[zonesindex]:
+                            groupzone_members[zonesindex].remove(addedzone)
+                    media_callback[zonesindex]()
+                groupzone_members[zoneindex].append(addedzone)
+                media_callback[entityindex]()
+                _LOGGER.debug('Join zone %s to controller zone %s.  All slave zones: %s', \
+                    addedzone, zone, groupzone_members[zoneindex])
+            callbackobj = media_zonetable.index(zone)
+            callbackint = int(callbackobj)
+            media_callback[callbackint]()
+
+        @synchronized
+        def unjoin_player(self, zone):
+            _LOGGER.debug('Unjoin zone %s from all groups', zone)
+            zoneindexobj = [i for i, x in enumerate(media_zonetable) if x == zone]
+            zoneindex = int(zoneindexobj[0])
+            if len(groupzone_members[zoneindex]) == 1:
+                self.set_power(zone, False)
+            for control_zone in media_zonetable:
+                control_zoneindexobj = \
+                    [i for i, x in enumerate(media_zonetable) if x == control_zone]
+                control_zoneindex = int(control_zoneindexobj[0])
+                if zone != control_zone:
+                    if zone in groupzone_members[control_zoneindex]:
+                        groupzone_members[control_zoneindex].remove(zone)
+                else:
+                    del groupzone_members[control_zoneindex][:]
+                    groupzone_members[control_zoneindex].append(media_zonetable[control_zoneindex])
+                callbackobj = media_zonetable.index(control_zone)
+                callbackint = int(callbackobj)
+                media_callback[callbackint]()
+
         @synchronized
         def set_power(self, zone: int, power: bool):
-            self._process_request(_format_set_power(zone, power))
+            if zone == 0:
+                if power:
+                    zonepwr[zone] = 1
+                else:
+                    zonepwr[zone] = 0
+                media_callback[0]()
+            else:
+                self._process_request(_format_set_power(zone, power))
 
         @synchronized
         def set_mute(self, zone: int, mute: bool):
-            self._process_request(_format_set_mute(zone, mute))
+            if zone == 0:
+                if mute:
+                    zonemute[0] = 1
+                else:
+                    zonemute[0] = 0
+                media_callback[0]()
+            else:
+                self._process_request(_format_set_mute(zone, mute))
 
         @synchronized
         def set_volume(self, zone: int, volume: float):
-            self._process_request(_format_set_volume(zone, volume))
+            if zone == 0:
+                zonevol[0] = volume
+                media_callback[0]()
+            else:
+                self._process_request(_format_set_volume(zone, volume))
+
+        @synchronized
+        def set_volume_offset(self, zone: int, volume: float):
+            zonevoloffset[zone] = volume
+            if zone in settings_zonetable:
+                callbackindex = [i for i, x in enumerate(settings_zonetable) if x == zone]
+                for callbackzone in callbackindex:
+                    settings_callback[callbackzone]()
 
         @synchronized
         def set_group(self, zone: int, group: bool):
@@ -612,7 +825,11 @@ def get_nuvo(port_url, baud):
 
         @synchronized
         def set_source(self, zone: int, source: int):
-            self._process_request(_format_set_source(zone, source))
+            if zone == 0:
+                zonesrc[0] = source
+                media_callback[0]()
+            else:
+                self._process_request(_format_set_source(zone, source))
 
         @synchronized
         def page_off(self, page_source, page_zones):
@@ -645,7 +862,7 @@ def get_nuvo(port_url, baud):
                    time.sleep(0.05)
                if zonesrc[int(page_zones[zonecnt])] != page_source:
                    self.set_source(int(page_zones[zonecnt]), page_source)
-               self.set_volume(int(page_zones[zonecnt]), int(page_volume[zonecnt]))
+               self.set_volume(int(page_zones[zonecnt]), percent_to_db(int(page_volume[zonecnt])))
                zonecnt += 1
 
         @synchronized
@@ -653,35 +870,30 @@ def get_nuvo(port_url, baud):
             self._process_request(_format_mute_all())
             for count in range(1,13):
                 zonemute[int(count)] = 1
-                try:
+                if count in media_zonetable:
                     callbackobj = media_zonetable.index(count)
                     callbackint = int(callbackobj)
                     media_callback[callbackint]()
-                except:
-                    callbackobj =  None
 
         @synchronized
         def unmute_all(self):
             self._process_request(_format_unmute_all())
             for count in range(1,13):
                 zonemute[int(count)] = 0
-                try:
+                if count in media_zonetable:
                     callbackobj = media_zonetable.index(count)
                     callbackint = int(callbackobj)
                     media_callback[callbackint]()
-                except:
-                    callbackobj =  None
 
         @synchronized
         def all_off(self):
             self._process_request(_format_all_off())
             for count in range(1,13):
+                zonepwralloff[int(count)] = 0
                 zonepwr[int(count)] = 0
-                try:
+                if count in media_zonetable:
                     callbackobj = media_zonetable.index(count)
                     callbackint = int(callbackobj)
                     media_callback[callbackint]()
-                except:
-                    callbackobj =  None
 
     return NuvoSync(port_url)
